package uk.ac.bbsrc.babraham.SparkSpotter;
/*
 * SparkSpotter - An ImageJ plugin for finding bright sparks in movies
 * Copyright (c) 2007 Simon Andrews

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import ij.ImageListener;
import ij.ImagePlus;
import ij.gui.Roi;
import ij.gui.StackWindow;

import javax.swing.ButtonGroup;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;


public class SparkDisplay extends JFrame implements ListSelectionListener, ImageListener {

	private SparkEventGroup [] events;
	private ImagePlus imagePlus;
	private SparkSpotterPreferences prefs;
	private JSplitPane splitPane;
	private SparkEventGroup selectedEvent = null;
	private GraphDisplay graph = new GraphDisplay();
	private JTable table;
	private int maxCorrected;
	private int minCorrected;
	private int maxRaw;
	private int minRaw;
	private String graphType = "corrected";
	private String scaleType = "global";
	private int currentFrame = 0;
	
	public SparkDisplay (SparkEventGroup [] events, ImagePlus imagePlus, SparkSpotterPreferences prefs) {
	
		this.events = events;
		this.imagePlus = imagePlus;
		this.prefs = prefs;
		ImagePlus.addImageListener(this);
		
		
		splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		this.setContentPane(splitPane);
		
		table = new JTable(new SparkTableModel());
		table.getSelectionModel().addListSelectionListener(this);
		// The method below was added in java 1.5, and isn't available in the old JRE
		// which is distributed with ImageJ.  We therefore need to catch a NoSuchMethodError
		try {
			table.setAutoCreateRowSorter(true);
		}
		catch (NoSuchMethodError e) {
			System.err.println("JRE is too old to support table sorting.  Disabling this feature.");
		};
		splitPane.setTopComponent(new JScrollPane(table));
		splitPane.setBottomComponent(graph);
		
		// Set the max & min values for the graph:
		for (int i=0;i<events.length;i++) {
			if (i==0) {
				maxCorrected = events[i].maxCorrected();
				minCorrected = events[i].minCorrected();
				maxRaw = events[i].maxRaw();
				minRaw = events[i].minRaw();
				continue;
			}
			if (events[i].maxCorrected() > maxCorrected) maxCorrected = events[i].maxCorrected();
			if (events[i].minCorrected() < minCorrected) minCorrected = events[i].minCorrected();
			if (events[i].maxRaw() > maxRaw) maxRaw = events[i].maxRaw();
			if (events[i].minRaw() < minRaw) minRaw = events[i].minRaw();
			
		}
		
		setTitle("SparkSpotter Results");
		setJMenuBar(new SparkDisplayMenuBar());
		setSize(600,600);
		setLocationRelativeTo(null);
		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		setVisible(true);
	}

	public void imageClosed(ImagePlus image) {
		if (image == imagePlus) {
			prefs.setVisible(false);
			prefs.dispose();

			setVisible(false);
			dispose();
		}
	}

	public void imageOpened(ImagePlus arg0) {
	}

	public void imageUpdated(ImagePlus arg0) {
		currentFrame = imagePlus.getCurrentSlice();
		graph.repaint();
	}

	public void exportTable () {
		
		JFileChooser chooser = new JFileChooser();
		chooser.setMultiSelectionEnabled(false);
		chooser.setFileFilter(new FileFilter() {
		
			public String getDescription() {
				return "Text files";
			}
		
			public boolean accept(File f) {
				if (f.isDirectory() || f.getName().toLowerCase().endsWith(".txt")) {
					return true;
				}
				else {
					return false;
				}
			}
		
		});
		
		int result = chooser.showSaveDialog(this);
		if (result == JFileChooser.CANCEL_OPTION) return;

		File file = chooser.getSelectedFile();
		if (! file.getPath().toLowerCase().endsWith(".txt")) {
			file = new File(file.getPath()+".txt");
		}

		// Check if we're stepping on anyone's toes...
		if (file.exists()) {
			int answer = JOptionPane.showOptionDialog(this,file.getName()+" exists.  Do you want to overwrite the existing file?","Overwrite file?",0,JOptionPane.QUESTION_MESSAGE,null,new String [] {"Overwrite and Save","Cancel"},"Overwrite and Save");

			if (answer > 0) {
				return;
			}
		}

		
		try {
			PrintWriter p = new PrintWriter(new FileWriter(file));
			
			TableModel model = table.getModel();
			
			int rowCount = model.getRowCount();
			int colCount = model.getColumnCount();
			
			// Do the headers first
			StringBuffer b = new StringBuffer();
			for (int c=0;c<colCount;c++) {
				b.append(model.getColumnName(c));
				if (c+1 != colCount) {
					b.append("\t");
				}
			}
			
			p.println(b);
			
			for (int r=0;r<rowCount;r++) {
				b = new StringBuffer();
				for (int c=0;c<colCount;c++) {
					b.append(model.getValueAt(r,c));
					if (c+1 != colCount) {
						b.append("\t");
					}
				}
				p.println(b);
				
			}
			p.close();
		
		}
			
				
		catch (FileNotFoundException e) {
			JOptionPane.showMessageDialog(this,"Can't save probe list: e.getMessage()","Error",JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
		}
		catch (IOException e) {
			JOptionPane.showMessageDialog(this,"Can't save probe list: e.getMessage()","Error",JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
		}
		
		
	}

	public void valueChanged(ListSelectionEvent e) {
		int [] i = table.getSelectedRows();
		if (i.length==1) {
			selectedEvent = events[table.convertRowIndexToModel(i[0])];
			((StackWindow)imagePlus.getWindow()).showSlice(selectedEvent.startFrame()+1);
			imagePlus.setRoi(new Roi(selectedEvent.getStartX()-(1+prefs.searchArea()),selectedEvent.getStartY()-(1+prefs.searchArea()),(selectedEvent.getEndX()-selectedEvent.getStartX())+3+(prefs.searchArea()*2),(selectedEvent.getEndY()-selectedEvent.getStartY())+3+(prefs.searchArea()*2)));
		}
		else {
			selectedEvent = null;
		}
		
		graph.repaint();
	}

	private class SparkTableModel extends AbstractTableModel {

		public int getColumnCount() {
			return 9;
		}

		public int getRowCount() {
			return events.length;
		}
		
		public String getColumnName (int columnIndex) {
			switch (columnIndex) {
			case 0:
				return "StartX";
			case 1:
				return "End X";
			case 2:
				return "StartY";
			case 3:
				return "EndY";
			case 4:
				return "Event Count";
			case 5:
				return "Start Frame";
			case 6:
				return "End Frame";
			case 7:
				return "Diff";
			case 8:
				return "Fold";
			default:
				return null;
			}
		}
		
		public Class<?> getColumnClass (int columnIndex) {
			switch (columnIndex) {
			case 0:
				return Integer.class;
			case 1:
				return Integer.class;
			case 2:
				return Integer.class;
			case 3:
				return Integer.class;
			case 4:
				return Integer.class;
			case 5:
				return Integer.class;
			case 6:
				return Integer.class;
			case 7:
				return Double.class;
			case 8:
				return Double.class;
			default:
				return null;
			}

		}

		public Object getValueAt(int rowIndex, int columnIndex) {
			switch (columnIndex) {
			case 0:
				return new Integer(events[rowIndex].getStartX());
			case 1:
				return new Integer(events[rowIndex].getEndX());
			case 2:
				return new Integer(events[rowIndex].getStartY());
			case 3:
				return new Integer(events[rowIndex].getEndY());
			case 4:
				return new Integer(events[rowIndex].eventCount());
			// The frames in SparkSpotter (and ImageJ) are 0 indexed, but the display
			// shows them as 1 indexed so we need to increment them in the table.
			case 5:
				return new Integer(events[rowIndex].startFrame()+1);
			case 6:
				return new Integer(events[rowIndex].endFrame()+1);
			case 7:
				return new Double(events[rowIndex].averageIntensity());
			case 8:
				return new Double(events[rowIndex].averageFold());
			default:
				return null;
			}
		}
	}
	
	private class GraphDisplay extends JPanel implements MouseListener, MouseMotionListener, MouseWheelListener {
		int graphMin;
		int graphMax;
		
		public GraphDisplay () {
			super();
			addMouseListener(this);
			addMouseMotionListener(this);
			addMouseWheelListener(this); 
		}
		
		public void paintComponent (Graphics g) {
			super.paintComponent(g);

			g.setColor(Color.WHITE);
			g.fillRect(0, 0, getWidth(), getHeight());
			
			if (selectedEvent == null) return;

			
			// Collect the data we're using for this repaint
			int [] data;
			if (graphType.equals("corrected")) {
				data = selectedEvent.correctedData();
				if (scaleType.equals("local")) {
					graphMin = selectedEvent.minCorrected();
					graphMax = selectedEvent.maxCorrected();
				}
				else {
					graphMin = minCorrected;
					graphMax = maxCorrected;
				}
			}
			else {
				data = selectedEvent.rawData();
				if (scaleType.equals("local")) {
					graphMin = selectedEvent.minRaw();
					graphMax = selectedEvent.maxRaw();
				}
				else {
					graphMin = minRaw;
					graphMax = maxRaw;
				}
			}
						
			// Draw the axes
			g.setColor(Color.BLACK);
			
			g.drawLine(20, 20, 20, getHeight()-20);
			g.drawLine(20, getHeight()-20, getWidth()-5, getHeight()-20);

			if (graphType.equals("corrected")) {
				g.setColor(Color.LIGHT_GRAY);
				g.drawLine(20, calculateY(0), getWidth()-5, calculateY(0));
			}
			
			g.setColor(Color.BLACK);
			g.drawString("Frame "+currentFrame+" of "+imagePlus.getStackSize(), (getWidth()/2)-25, getHeight()-5);
			g.drawString("I", 5, getHeight()/2);
			
			g.drawString("X="+selectedEvent.getStartX()+"-"+selectedEvent.getEndX()+" Y="+selectedEvent.getStartY()+"-"+selectedEvent.getEndY()+" Frames "+selectedEvent.startFrame()+"-"+selectedEvent.endFrame(), 15, 15);

			// Now we iterate through the data plotting the points
			

			
			int lastX=0;
			int lastY=0;
			int thisX;
			int thisY;
			
//			System.err.println("We have data for "+data.length+" frames");
			for (int f=0;f<data.length;f++) {
				if (f >= selectedEvent.startFrame() && f <= selectedEvent.endFrame()+1) {
					g.setColor(Color.RED);
				}
				else {
					g.setColor(Color.BLACK);
				}
//				System.err.println("Plotting frame "+f+" with value "+data[f]);
				thisX = calculateX(f, data.length-1);
				thisY = calculateY(data[f]);
				if (f>0) {
//					System.err.println("Drawing "+lastX+","+lastY+","+thisX+","+thisY);

					g.drawLine(lastX, lastY, thisX, thisY);
				}
				lastX = thisX;
				lastY = thisY;
			}
			
			// Add on a line indicating the current frame
			g.setColor(Color.BLUE );
			g.drawLine(calculateX(currentFrame-1,data.length-1), 20, calculateX(currentFrame-1,data.length-1), getHeight()-20);
			
			
		}
		
		private int calculateX (int x, int max) {
			// Scales between 20 and width-5 and starts at 0
			return (int)(20+((getWidth()-25))*(double)x/max);
		}
		
		private int calculateFromX (int x, int max) {
			// Takes an x coordinate from the window and works out the frame
			return 1+(int)(((x-20d)/(getWidth()-25))*max);			
		}
		
		private int calculateY (int y) {
			return (getHeight()-20)-(int)((getHeight()-40)*(((double)y-graphMin)/(graphMax-graphMin)));			
		}

		public void mouseClicked(MouseEvent e) {}

		public void mouseEntered(MouseEvent e) {}

		public void mouseExited(MouseEvent e) {}

		public void mousePressed(MouseEvent e) {
			int f = calculateFromX(e.getX(), imagePlus.getStackSize());
			
			((StackWindow)imagePlus.getWindow()).showSlice(f);
		
		}

		public void mouseReleased(MouseEvent e) {}

		public void mouseDragged(MouseEvent e) {
			int f = calculateFromX(e.getX(), imagePlus.getStackSize());
			
			((StackWindow)imagePlus.getWindow()).showSlice(f);
		}

		public void mouseMoved(MouseEvent e) {}

		public void mouseWheelMoved(MouseWheelEvent e) {
			int w = e.getWheelRotation();
			int f = currentFrame+w;
			if (f<1) f=1;
			if (f>imagePlus.getStackSize()) f=imagePlus.getStackSize();
			((StackWindow)imagePlus.getWindow()).showSlice(f);
			
		}
	}

	private class SparkDisplayMenuBar extends JMenuBar implements ActionListener {
		
		public SparkDisplayMenuBar () {
			JMenu fileMenu = new JMenu("File");
			fileMenu.setMnemonic(KeyEvent.VK_F);
			
			JMenuItem fileSave = new JMenuItem("Save");
			fileSave.setMnemonic(KeyEvent.VK_C);
			fileSave.setActionCommand("save");
			fileSave.addActionListener(this);
			fileMenu.add(fileSave);

			fileMenu.addSeparator();
			
			JMenuItem fileClose = new JMenuItem("Close");
			fileClose.setMnemonic(KeyEvent.VK_C);
			fileClose.setActionCommand("close");
			fileClose.addActionListener(this);
			fileMenu.add(fileClose);
			
			add(fileMenu);
			
			JMenu displayMenu = new JMenu("Display");
			displayMenu.setMnemonic(KeyEvent.VK_D);
			
			ButtonGroup dataSources = new ButtonGroup();
			JRadioButtonMenuItem corrected = new JRadioButtonMenuItem("Corrected Data",true);
			corrected.addActionListener(this);
			corrected.setActionCommand("corrected");
			dataSources.add(corrected);
			displayMenu.add(corrected);
			JRadioButtonMenuItem raw = new JRadioButtonMenuItem("Raw");
			raw.addActionListener(this);
			raw.setActionCommand("raw");
			dataSources.add(raw);
			displayMenu.add(raw);
			displayMenu.addSeparator();
			
			ButtonGroup scaleOptions = new ButtonGroup();
			JRadioButtonMenuItem globalScale = new JRadioButtonMenuItem("Scale Globally",true);
			globalScale.setActionCommand("global");
			globalScale.addActionListener(this);
			scaleOptions.add(globalScale);
			displayMenu.add(globalScale);
			JRadioButtonMenuItem localScale = new JRadioButtonMenuItem("Scale Locally");
			localScale.setActionCommand("local");
			localScale.addActionListener(this);
			scaleOptions.add(localScale);
			displayMenu.add(localScale);
			
			add(displayMenu);
			
		}

		public void actionPerformed(ActionEvent e) {
			if (e.getActionCommand().equals("close")) {
				setVisible(false);
				dispose();
			}
			else if (e.getActionCommand().equals("save")) {
				exportTable();
			}
			else if (e.getActionCommand().equals("corrected")) {
				graphType = "corrected";
				graph.repaint();
			}
			else if (e.getActionCommand().equals("raw")) {
				graphType = "raw";
				graph.repaint();
			}
			else if (e.getActionCommand().equals("local")) {
				scaleType = "local";
				graph.repaint();
			}
			else if (e.getActionCommand().equals("global")) {
				scaleType = "global";
				graph.repaint();
			}
		}
	}
}
